/*
 * Copyright 2019 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc.xds.sds;

import static com.google.common.base.Preconditions.checkNotNull;

import io.envoyproxy.envoy.api.v2.auth.CommonTlsContext;

/** Class for utility functions for {@link CommonTlsContext}. */
final class CommonTlsContextUtil {

  private CommonTlsContextUtil() {}

  /** Returns true only if given CommonTlsContext uses no SdsSecretConfigs. */
  static boolean hasAllSecretsUsingFilename(CommonTlsContext commonTlsContext) {
    checkNotNull(commonTlsContext, "commonTlsContext");
    // return true if it has no SdsSecretConfig(s)
    return (commonTlsContext.getTlsCertificateSdsSecretConfigsCount() == 0)
        && !commonTlsContext.hasValidationContextSdsSecretConfig();
  }

  /** Returns true only if given CommonTlsContext uses only SdsSecretConfigs. */
  static boolean hasAllSecretsUsingSds(CommonTlsContext commonTlsContext) {
    checkNotNull(commonTlsContext, "commonTlsContext");
    // return true if it has only SdsSecretConfig(s)
    return (commonTlsContext.getTlsCertificatesCount() == 0)
        && !commonTlsContext.hasValidationContext();
  }
}
