// Copyright 2018 The gRPC Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// The canonical version of this proto can be found at
// https://github.com/grpc/grpc-proto/blob/master/grpc/gcp/altscontext.proto

syntax = "proto3";

package grpc.gcp;

import "grpc/gcp/transport_security_common.proto";

option go_package = "google.golang.org/grpc/credentials/alts/internal/proto/grpc_gcp";
option java_multiple_files = true;
option java_outer_classname = "AltsContextProto";
option java_package = "io.grpc.alts.internal";

message AltsContext {
  // The application protocol negotiated for this connection.
  string application_protocol = 1;

  // The record protocol negotiated for this connection.
  string record_protocol = 2;

  // The security level of the created secure channel.
  SecurityLevel security_level = 3;

  // The peer service account.
  string peer_service_account = 4;

  // The local service account.
  string local_service_account = 5;

  // The RPC protocol versions supported by the peer.
  RpcProtocolVersions peer_rpc_versions = 6;

  // Additional attributes of the peer.
  map<string, string> peer_attributes = 7;
}
