/*
 * Copyright 2015 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc.okhttp;

import io.grpc.internal.WritableBuffer;
import io.grpc.internal.WritableBufferTestBase;
import okio.Buffer;
import org.junit.Before;
import org.junit.runner.RunWith;
import org.junit.runners.JUnit4;

/**
 * Tests for {@link OkHttpWritableBuffer}.
 */
@RunWith(JUnit4.class)
public class OkHttpWritableBufferTest extends WritableBufferTestBase {

  private OkHttpWritableBuffer buffer;

  @Before
  public void setup() {
    buffer = new OkHttpWritableBuffer(new Buffer(), 100);
  }

  @Override
  protected WritableBuffer buffer() {
    return buffer;
  }

  @Override
  protected byte[] writtenBytes() {
    return buffer.buffer().readByteArray();
  }
}
