/*
 * Copyright 2015 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc;

import java.io.InputStream;

/** Marshalls decimal-encoded integers. */
public class IntegerMarshaller implements MethodDescriptor.Marshaller<Integer> {
  public static final IntegerMarshaller INSTANCE = new IntegerMarshaller();

  @Override
  public InputStream stream(Integer value) {
    return StringMarshaller.INSTANCE.stream(value.toString());
  }

  @Override
  public Integer parse(InputStream stream) {
    return Integer.valueOf(StringMarshaller.INSTANCE.parse(stream));
  }
}
