// Copyright 2015 The gRPC Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
syntax = "proto3";

package grpc.testing;

option java_package = "io.grpc.benchmarks.proto";
option java_outer_classname = "Payloads";

message ByteBufferParams {
  int32 req_size = 1;
  int32 resp_size = 2;
}

message SimpleProtoParams {
  int32 req_size = 1;
  int32 resp_size = 2;
}

message ComplexProtoParams {
  // TODO (vpai): Fill this in once the details of complex, representative
  //              protos are decided
}

message PayloadConfig {
  oneof payload {
    ByteBufferParams bytebuf_params = 1;
    SimpleProtoParams simple_params = 2;
    ComplexProtoParams complex_params = 3;
  }
}
