/*
 * Copyright 2018 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc.grpclb;

import java.net.SocketAddress;

final class FakeSocketAddress extends SocketAddress {
  final String name;

  FakeSocketAddress(String name) {
    this.name = name;
  }

  @Override
  public String toString() {
    return "FakeSocketAddress-" + name;
  }

  @Override
  public boolean equals(Object other) {
    if (other instanceof FakeSocketAddress) {
      FakeSocketAddress otherAddr = (FakeSocketAddress) other;
      return name.equals(otherAddr.name);
    }
    return false;
  }

  @Override
  public int hashCode() {
    return name.hashCode();
  }
}
