/*
 * Copyright 2015 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc.netty;

import io.grpc.internal.ClientTransport.PingCallback;
import java.util.concurrent.Executor;

/**
 * Command sent from the transport to the Netty channel to send a PING frame.
 */
class SendPingCommand extends WriteQueue.AbstractQueuedCommand {
  private final PingCallback callback;
  private final Executor executor;

  SendPingCommand(PingCallback callback, Executor executor) {
    this.callback = callback;
    this.executor = executor;
  }

  PingCallback callback() {
    return callback;
  }

  Executor executor() {
    return executor;
  }
}
