/*
 * Copyright 2016 The gRPC Authors
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package io.grpc.benchmarks;

import io.grpc.Drainable;
import io.grpc.KnownLength;
import io.netty.buffer.ByteBuf;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

/**
 * A {@link Drainable} {@code InputStream} that reads an {@link ByteBuf}.
 */
@SuppressWarnings("InputStreamSlowMultibyteRead") // doesn't matter if slow. It'll throw
public class ByteBufInputStream extends InputStream
    implements Drainable, KnownLength {

  private ByteBuf buf;

  ByteBufInputStream(ByteBuf buf) {
    this.buf = buf;
  }

  @Override
  public int drainTo(OutputStream target) throws IOException {
    int readableBytes = buf.readableBytes();
    buf.readBytes(target, readableBytes);
    buf = null;
    return readableBytes;
  }

  @Override
  public int available() throws IOException {
    if (buf != null) {
      return buf.readableBytes();
    }
    return 0;
  }

  @Override
  public int read() throws IOException {
    throw new UnsupportedOperationException();
  }
}
