// Copyright 2016 The gRPC Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// Service definition designed to test the ProtoReflectionService when a service
// is dynamically added and removed.

syntax = "proto2";

import "io/grpc/reflection/testing/dynamic_reflection_test_depth_two.proto";

option java_multiple_files = true;
option java_package = "io.grpc.reflection.testing";
option java_outer_classname = "DynamicReflectionTestProto";

package grpc.reflection.testing;

// A DynamicService
service DynamicService {
  // A method
  rpc Method (DynamicRequest) returns (DynamicReply) {}
}

// AnotherDynamicService
service AnotherDynamicService {
  // A method
  rpc Method (DynamicRequest) returns (DynamicReply) {}
}
